'**********************************************************************************************
'*  io.spin - handle input, output and break requests sent by qz80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id$
'**********************************************************************************************
CON

' Define this to check for contiguous image files
#define	CHECK_CONTIG

' Define this to make 128 byte record writes pending for 1 second
#define TIMED_FLUSH

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read (TRS-80 keyboard)

	constat_port		=	$10		' SIO port 0 status port
	condata_port		=	$11		' SIO port 0 data port
	punstat_port		=	$12		' punch 1 status port
	pundata_port		=	$13		' punch/reader 1 data port
	sio1stat_port		=	$14		' SIO port 1 status port
	sio1data_port		=	$15		' SIO port 1 data port
	sio2stat_port		=	$16		' SIO port 2 status port
	sio2data_port		=	$17		' SIO port 2 data port
	sio3stat_port		=	$18		' SIO port 3 status port
	sio3data_port		=	$19		' SIO port 3 data port
	lcdstat_port		=	$1a		' LCD status port
	lcddata_port		=	$1b		' LCD data port

	keystat_port		=	$74		' test just keyboard status (constat does this, too)
	keydata_port		=	$75		' get just the keyboard character

	hdsk_reset		=	$1		' command to reset controller
	hdsk_read		=	$2		' read command
	hdsk_write		=	$3		' write command
	hdsk_param		=	$4		' get hard disk parameters
	hdsk_port		=	$fd		' control port for simulated hard disk

	simh_set_bank_select	=	12		' SIMH command to set memory bank
	simh_get_common		=	13		' SIMH get common memory base
	simh_reset		=	14		' SIMH reset command
	simh_has_banked_memory	=	18		' SIMH command to check for banked memory
	simh_start_timer		=	21		' SIMH start timer interrupt
	simh_stop_timer		=	22		' SIMH stop timer interrupt
	simh_timer_delta		=	23		' SIMH set timer delta in milliseconds
	simh_timer_addr		=	24		' SIMH set timer interrupt address
	simh_port		=	$fe		' control port for simulated hard disk

	baud			=	115200		' baud rate

	dir_none		=	0		' no i/o direction selected
	dir_out			=	1		' expecting client to write to the port
	dir_in			=	2		' expecting client to read from the port

	lmm_mul32		=	0		' multiply x by y
	lmm_div32		=	1		' divide x by y
	lmm_break		=	2		' function number for break
	lmm_keystat		=	3		' status of keyboard input
	lmm_keydata		=	4		' data from keyboard input
	lmm_st0			=	5		' function number for st0 - status of 1st serial port input
	lmm_tx0			=	6		' function number for tx0 - send to 1st serial port
	lmm_rx0			=	7		' function number for rx0 - receive from 1st serial port
	lmm_st1			=	8		' function number for st1 - status of 2nd serial port input
	lmm_tx1			=	9		' function number for tx1 - send to 2nd serial port
	lmm_rx1			=	10		' function number for rx1 - receive from 2nd serial port
	lmm_vt100_out		=	11		' function number for VT100 output
	lmm_lcd_init		=	12		' function number for LCD initialization

#ifdef HAVE_TV
	SCREEN_W		=	80
	SCREEN_H		=	25
#endif HAVE_TV
#ifdef HAVE_VGA
	SCREEN_W		=	80
	SCREEN_H		=	40
#endif HAVE_VGA

VAR
	long	cog
	long	params[8]
	long	io_command				' io command
	long	bankbase				' banked memory base address
#ifdef HAVE_TV
	long	vt_command
#endif HAVE_TV
#ifdef HAVE_VGA
	long	vt_command
#endif HAVE_VGA
	byte	name[12+1]
	byte	datetime[20+1]

OBJ
	fat	: "fatfs"
	sio	: "pcFullDuplexSerial4FC"
#ifdef HAVE_KBD
	kbd	: "Keyboard"
#endif HAVE_KBD
#ifdef HAVE_TV
	tv80	: "tv80"
	vt100	: "vt100"
#endif
#ifdef HAVE_VGA
	vga	: "VGA_HiRes_Text"
	vt100	: "vt100"
#endif HAVE_VGA

PUB start(pins,irq) : okay | i, dent, sec, size
	stop

	irq_ptr := irq

	' This object has to be first, or it just won't work... talking about taking speed to the limit!?
	okay := \fat.start(iobuff_ptr, byte[pins][4], byte[pins][5], byte[pins][6], byte[pins][7])
	if okay < 0
		return -1
	pdo_command := fat.get_SPI_command
	pdo_blocknr := pdo_command + 4
	pdo_address := pdo_command + 8

	sio.AddPort(0, byte[pins][0], byte[pins][1], -1, -1, 0, 0, baud)
	if byte[pins][2] <> $ff
		sio.AddPort(1, byte[pins][2], byte[pins][3], -1, -1, 0, 0, baud)

	okay := sio.Start
	if okay =< 0
		return
	success(string("SIO"))
	tx_head0 := sio.get_tx_head
	tx_tail0 := sio.get_tx_tail
	tx_buffer0 := sio.get_tx_buffer
	rx_head0 := sio.get_rx_head
	rx_tail0 := sio.get_rx_tail
	rx_buffer0 := sio.get_rx_buffer

	if byte[pins][2] <> $ff
		tx_head1 := sio.get_tx_head1
		tx_tail1 := sio.get_tx_tail1
		tx_buffer1 := sio.get_tx_buffer1
		rx_head1 := sio.get_rx_head1
		rx_tail1 := sio.get_rx_tail1
		rx_buffer1 := sio.get_rx_buffer1

#ifdef HAVE_KBD
	if byte[pins][8] == $ff or byte[pins][9] == $ff
		return -1
	okay := \kbd.startx(byte[pins][8], byte[pins][9], %100, 40)
	if okay < 0
		failure(string("KBD"),okay)
		return -1
	success(string("KBD"))
	kbd_tail := kbd.tail_ptr
	kbd_head := kbd.head_ptr
	kbd_keys := kbd.keys_ptr
#endif

#ifdef HAVE_VGA
	vt_cmdptr := @vt_command
	if byte[pins][10] == $ff
		return -1
	okay := \vga.start(byte[pins][10], @vga_screen, @vga_colors, @vga_cursor, @vga_vsync)
	if okay <> TRUE
		return -1
	wordfill(@vga_colors, %%0020_3300, 64)
	vga_cursor[0] := 0		' cursor x
	vga_cursor[1] := 0		' cursor y
	vga_cursor[2] := %110		' on, blink slow
#endif HAVE_VGA

#ifdef HAVE_TV
	if byte[pins][11] == $ff
		return -1
	params[0] := byte[pins][11]
	params[1] := @tv_screen		' TV screen buffer
	okay := tv80.start(@params)
	if okay =< 0
		failure(string("TV 80"),okay)
		return
	success(string("TV80"))
	vt_cmdptr := @vt_command
#endif HAVE_TV

	str(string("qZ80 I/O starting...",13,10))

	okay := \fat.mount
	if okay < 0
		failure(string("FAT mount"),okay)
		return
	str(string("Volume serial #"))
	hex(\fat.get_serialno >> 16, 4)
	chr("-")
	hex(\fat.get_serialno, 4)
	str(string(", label "))
	str(\fat.get_volume_label)
	crlf

	read_boot(@boot_name)
	repeat i from 0 to 7
		map_disk(@lba_drive_a + i << 2, drive_name[i])

	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))


#ifdef HAVE_VGA
	params[0] := @vt_command
	params[1] := @vga_screen		' VGA screen buffer
	params[2] := @vga_colors		' VGA row colors
	params[3] := @vga_cursor		' VGA cursor positions
	params[4] := @vga_vsync			' VGA vertical sync flag
	params[5] := SCREEN_W			' terminal width
	params[6] := SCREEN_H			' terminal height
	okay := \vt100.start(@params)
	if okay =< 0
		failure(string("VT100 emulation"),okay)
		return
#endif

#ifdef HAVE_TV
	params[0] := @vt_command
	params[1] := @tv_screen			' TV screen buffer
	params[2] := 0				' no colors
	params[3] := 0				' no cursor
	params[4] := 0				' no vsync
	params[5] := SCREEN_W			' terminal width
	params[6] := SCREEN_H			' terminal height
	okay := \vt100.start(@params)
	if okay =< 0
		failure(string("VT100"),okay)
		return
	success(string("VT100"))
#endif
	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PUB command_addr
'' Return the address of the 3 longs block
	return @io_command

PUB chr(ch)
	sio.tx(0,ch)

PUB str(str_ptr) | i
	repeat i from 0 to strsize(str_ptr)-1
		chr(byte[str_ptr][i])

PUB crlf
	chr(13)
	chr(10)

PUB dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PUB dec(value) | _i
'' Print a decimal number
	if value < 0
		-value
		chr("-")

	_i := 1_000_000_000

	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PUB hex(value, digits)
'' Print a hexadecimal number
	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number
	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who)
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI read_boot(fname_ptr) | dent, size, sec
	dent := fat.find_file(fname_ptr)
	if dent == 0
		failure(fname_ptr,dent)
		return -1
	size := fat.get_dent_size(dent)
	sec := fat.get_dent_first_sector(dent)
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(", sector "))
	hex(sec,6)
	str(string(", size "))
	dec_size(size)
	str(string(", "))
	str(@datetime)
	crlf
	fat.read_sector(sec)
	fat.write_xmm_ram($FF00, 256)
	long[iobuff_ptr] := $00FF00C3
	fat.write_xmm_ram($0000, 4)


PRI map_disk(lbaptr, fname_ptr) : okay | dent, size, clus, exp, err, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		failure(fname_ptr, dent)
		return -1
	sec := fat.get_dent_first_sector(dent)
	size := fat.get_dent_size(dent)
	long[lbaptr] := sec
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(", sector "))
	hex(sec, 6)
	str(string(", size "))
	dec_size(size)
	str(string(", "))
	str(@datetime)
#ifdef CHECK_CONTIG
	clus := fat.get_dent_cluster(dent)
	exp := clus + 1
	err := 0
	repeat while (clus := fat.fat_next(clus)) => 0
		if clus <> exp
			err++
			quit
		exp := exp + 1
	if err
		str(string(" not contiguous!"))
	else
		str(string(" contiguous - okay."))
#endif
	crlf

DAT
'---------------------------------------------------------------------------------------------------------
		org	0
'---------------------------------------------------------------------------------------------------------
#ifdef HAVE_VGA
vga_screen	long
#endif HAVE_VGA
#ifdef HAVE_TV
tv_screen	long
#endif HAVE_TV
'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr	mov	iocmd_ptr, PAR
iocmd		mov	iobuff_lba, minus1
#ifdef BANKED_MEM
bankbase_ptr	mov	bankbase_ptr, PAR
		add	bankbase_ptr, #4
#endif
		mov	timer_tick, CNT
		sub	timer_tick, #5
		jmp	#startup

' pcFullDuplexSerial4FC
tx_head0	long	0-0
tx_tail0	long	0-0
tx_buffer0	long	0-0
rx_head0	long	0-0
rx_tail0	long	0-0
rx_buffer0	long	0-0

tx_head1	long	0-0
tx_tail1	long	0-0
tx_buffer1	long	0-0
rx_head1	long	0-0
rx_tail1	long	0-0
rx_buffer1	long	0-0

sio_head	long	0
sio_tail	long	0
sio_temp	long	0

#ifdef HAVE_KBD
kbd_tail	long	0-0
kbd_head	long	0-0
kbd_keys	long	0-0
kbd_ctrl	long	1<<9
#endif HAVE_KBD

#ifdef HAVE_TV
vt_cmdptr	long	0-0
#endif

#ifdef HAVE_VGA
vt_cmdptr	long	0-0
#endif

pdo_command	long	0-0
pdo_blocknr	long	0-0					' SD block number for r/w commands
pdo_address	long	0-0					' hub RAM address for r/w commands
iobuff_ptr	long	@@@iobuff

lmm_func_ptr	long	@@@lmm_func_table
lmm_pc		long	0-0

irq_ptr		long	0-0					' set by Spin code
timer_tick	long	0					' tick timer count
timer_delta	long	20*80_000				' timer delta (clock cycles for milliseconds)
timer_addr	long	$0038cd					' timer interrupt address

minus1		long	$ffffffff
low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
h128		long	128<<16
i1000		long	1000

msg_break_ptr	long	@@@msg_break

#ifdef HAVE_LCD
latch_dir	long	%00000000_00000000_00001111_11111111	' 138 active, gate active and 8 data lines active
latch_lcd	long	%00000000_00000000_00001010_00000000	' LCD latch is xxxx101x and gate low xxxxxxx0
enable_lcd	long	%00000000_00000000_00000000_00000010	' Enable pin 6 on LCD displays
lcd_delay	long	100_000
lcd_or		long	0
#endif

addr		long	0
port		long	0
data		long	0
io_dir		long	0

#ifdef BANKED_MEM
bank		long	0					' current RAM bank base address
banklimit	long	$b000					' the limit for the banked memory range
bank64k		long	64*1024					' 64K from bank #0 to bank #1
bankmax		long	512*1024
#endif

output_port_ptr	long	@@@output_port_table			' list of output ports and corresponding function
input_port_ptr	long	@@@input_port_table			' list of input ports and corresponding function

simh_cmd_ptr	long	@@@simh_cmd_table
simh_cmd	long	0					' currently active command
simh_status	long	0					' status byte
simh_timer	long	0					' timer enable status
simh_params	long	@@@simh_params_				' pointer to the simh io buffer
simh_params_ptr	long	0					' pointer into simh_params
simh_params_cnt	long	0					' count of parameters expected

hdsk_cmd	long	0
hdsk_status	long	0
hdsk_lba_drives	long	@@@lba_drive_a
hdsk_drive	long	@@@hdsk_drive_				' parameters of read/write command
hdsk_sector	long	@@@hdsk_sector_
hdsk_track	long	@@@hdsk_track_
hdsk_dma	long	@@@hdsk_dma_
hdsk_dpb	long	@@@hdsk_dpb_
hdsk_params_ptr	long	0					' pointer into hdsk_params in hub RAM
hdsk_params_cnt	long	0					' count of parameters expected

lba0		long	0-0					' logical block address of CP/M
lba		long	0					' logical block translated from track*32+sector
iobuff_lba	long	-1					' logical block that is currently in iobuff
sector_offs	long	0					' offset of current sector in iobuff_ptr
#ifdef TIMED_FLUSH
time_dirty	long	0
#endif TIMED_FLUSH

t1		long	0
t2		long	0
t3		long	0
x		long	0
y		long	0

ptr		long	0
hexval		long	0
hexcnt		long	0

startup
#ifdef HAVE_LCD
		mov	lmm_pc, #lmm_lcd_init
		call	#lmm_func
#endif
		mov	data, #12				' form feed
		jmp	#output_conout

command

#ifdef TIMED_FLUSH
		mov	t1, time_dirty		WZ
	if_nz	sub	t1, CNT
	if_nz	cmps	t1, #0			WC
  if_nz_and_c	call	#spi_write_block			' yes, write the buffer and clear the dirty flag
#endif TIMED_FLUSH

#ifdef INTERRUPTS
		tjz	simh_timer, #:ints_off			' interrupts turned off?
		mov	t1, timer_tick
		sub	t1, CNT
		cmps	t1, #0			WC
	if_c	call	#timer_int
:ints_off
#endif
		rdlong	iocmd, iocmd_ptr
		mov	t1, iocmd
		shr	t1, #24			WZ		' inspect bits 31..24: zero?
	if_z	jmp	#command				' yes, loop waiting for a command.

		mov	port, iocmd				' extract port from bits 23..8
		shr	port, #8				' to bits 15..0
		and	port, #$ff				' mask bits 7..0 of port address

		cmp	t1, #io_cmd_out		WZ
	if_z	jmp	#output
		cmp	t1, #io_cmd_in		WZ
	if_z	jmp	#input
		cmp	t1, #io_cmd_break	WZ
	if_z	mov	lmm_pc, #lmm_break			' LMM function "break"
	if_z	call	#lmm_func
command_0
		and	iocmd, low_24bits			' clear command byte in bits 31..24
		wrlong	iocmd, iocmd_ptr
		jmp	#command

'**********************************************************************************************
'
' Read i/o port and data to write to
'
output
		mov	data, iocmd				' extract data
		and	data, #$ff				' mask 8 data bits
		mov	ptr, output_port_ptr
:loop		rdword	t1, ptr
		add	ptr, #2
		cmp	port, t1		WZ
	if_z	rdword	t1, ptr
	if_z	jmp	t1
		add	ptr, #2
		tjnz	t1, #:loop
		jmp	#command_0

'**********************************************************************************************
'
' Write data to the console. Writes to the FullDuplexSerial tx_buffer0
'
output_conout
		call	#conout_char
		jmp	#command_0

'**********************************************************************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'**********************************************************************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'**********************************************************************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	data, hexval
		shr	data, #28
		cmp	data, #10	WZ, WC
	if_b	add	data, #48
	if_ae	add	data, #55
		call	#conout_char
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'**********************************************************************************************
'
' Output a space character to the console
'
conout_space
		mov	data, #32
		call	#conout_char
conout_space_ret
		ret

'**********************************************************************************************
'
' Output a carriage return to the console
'
conout_crlf
		mov	data, #13
		call	#conout_char
		mov	data, #10
		call	#conout_char
conout_crlf_ret
		ret

'**********************************************************************************************
'
' Output a string at ptr to the console
'
conout_string
		rdbyte	data, ptr	WZ
		add	ptr, #1
	if_z	jmp	#conout_string_ret
		call	#conout_char
		jmp	#conout_string
conout_string_ret
		ret

'**********************************************************************************************
'
' Output to the punch - SIO port #1
'
output_pundata
		mov	lmm_pc, #lmm_tx1
		call	#lmm_func
		jmp	#command_0

'**********************************************************************************************
'
' Output an instruction code to the LCD
'
output_lcdctrl
		call	#lcd_send_inst
		jmp	#command_0

'**********************************************************************************************
'
' Output a text character to the LCD
'
output_lcddata
		call	#lcd_send_text
		jmp	#command_0

'**********************************************************************************************
'
' Handle SIMH compatible command
'
output_simh
		cmp	io_dir, #dir_out		WZ
	if_z	tjnz	simh_params_ptr, #output_simh_params
		mov	ptr, simh_cmd_ptr
:loop		rdword	t1, ptr
		add	ptr, #2
		cmp	data, t1			WZ
	if_z	rdword	t1, ptr
	if_z	jmp	t1
		add	ptr, #2
		tjnz	t1, #:loop
		mov	simh_status, #$01			' set error status
		jmp	#command_0

output_simh_params
		wrbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		sub	simh_params_cnt, #1		WZ
	if_nz	jmp	#command_0
		mov	simh_params_ptr, #0			' stop recording parameters
		' handle simh_cmd
		cmp	simh_cmd, #simh_set_bank_select WZ
	if_z	jmp	#output_simh_set_bank
		cmp	simh_cmd, #simh_timer_delta	WZ
	if_z	jmp	#output_simh_set_timerdelta
		cmp	simh_cmd, #simh_timer_addr	WZ
	if_z	jmp	#output_simh_set_timeraddr
		jmp	#command_0

output_simh_set_bank_select
		mov	simh_cmd, data
		mov	simh_params_ptr, simh_params
		mov	simh_params_cnt, #1			' expect one more output: the bank number
		mov	io_dir, #dir_out
		jmp	#command_0

output_simh_get_common
		mov	simh_cmd, data
		mov	simh_params_ptr, simh_params
		mov	simh_params_cnt, #2			' two bytes read on input
#ifdef BANKED_MEM
		wrword	banklimit, simh_params
#endif
		mov	io_dir, #dir_in
		jmp	#command_0

output_simh_reset
		mov	simh_params_ptr, #0			' stop recording or replying parameters
		mov	simh_params_cnt, #0
		mov	simh_cmd, data
output_simh_return
		mov	simh_status, #0
		mov	io_dir, #dir_none
		jmp	#command_0

output_simh_has_banked_memory
		mov	simh_cmd, data
#ifdef BANKED_MEM
		mov	simh_status, #10			' 10 banks (or 11?)
#else
		mov	simh_status, #0
#endif
		jmp	#command_0

output_simh_start_timer
		mov	simh_timer, #1
		jmp	#output_simh_return

output_simh_stop_timer
		mov	simh_timer, #0
		jmp	#output_simh_return

output_simh_timer_delta
output_simh_timer_addr
		mov	simh_cmd, data
		mov	simh_params_ptr, simh_params
		mov	simh_params_cnt, #2			' expect two bytes output
		mov	io_dir, #dir_out
		jmp	#command_0

output_simh_set_bank
#ifdef BANKED_MEM
		rdbyte	data, simh_params			' get bank number
		mov	bank, #0
		cmp	data, #0		WZ
	if_z	jmp	#:exit
		add	bank, bank64k
:loop		sub	data, #1		WZ		' decrement bank number
	if_z	jmp	#:exit
		add	bank, banklimit
		mov	t1, bank
		add	t1, banklimit
		cmp	t1, bankmax		WZ, WC
	if_ae	jmp	#command_0				' abort if end is beyond max bank
		jmp	#:loop
:exit		wrlong	bank, bankbase_ptr
#endif
		jmp	#output_simh_return

output_simh_set_timerdelta
		rdlong	x, #0					' get clkfreq
		mov	y, i1000				' divide by 1000
		mov	lmm_pc, #lmm_div32
		call	#lmm_func
		rdword	x, simh_params				' get delta in milliseconds
		mov	lmm_pc, #lmm_mul32
		call	#lmm_func
		mov	timer_delta, x				' delta in cycles
		jmp	#output_simh_return

output_simh_set_timeraddr
		rdword	timer_addr, simh_params
		shl	timer_addr, #8
		or	timer_addr, #$cd			' create a CALL $xxxx command
		jmp	#output_simh_return

'**********************************************************************************************
'
' Handle hard disk command
'
output_hdsk
		cmp	io_dir, #dir_out	WZ
	if_z	tjnz	hdsk_params_ptr, #output_hdsk_params

		cmp	data, #hdsk_reset	WZ
	if_z	jmp	#output_hdsk_reset
		cmp	data, #hdsk_read	WZ
	if_z	jmp	#output_hdsk_read
		cmp	data, #hdsk_write	WZ
	if_z	jmp	#output_hdsk_write
		cmp	data, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_drive
		mov	hdsk_status, #$01			' set error status?!
		jmp	#command_0

'**********************************************************************************************
'
' Write port value to RAM at hdsk_params_ptr
'
output_hdsk_params
		wrbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		sub	hdsk_params_cnt, #1 WZ
	if_nz	jmp	#command_0

		mov	hdsk_params_ptr, #0			' stop recording parameters
		cmp	hdsk_cmd, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_params_drive
		cmp	hdsk_cmd, #hdsk_read	WZ
	if_z	jmp	#hdsk_read_dma
		cmp	hdsk_cmd, #hdsk_write	WZ
	if_z	jmp	#hdsk_write_dma
		jmp	#command_0

'**********************************************************************************************
'
' Reset the hdsk interface
'
output_hdsk_reset
		mov	hdsk_cmd, data
		mov	hdsk_params_ptr, #0			' stop recording parameters
		mov	hdsk_params_cnt, #0
		mov	hdsk_status, #0
		mov	io_dir, #dir_none			' no direction
		jmp	#command_0

'**********************************************************************************************
'
' Read and write expect 6 additional
' writes to the port
'
output_hdsk_read
output_hdsk_write
		mov	hdsk_cmd, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #6			' drive, sector, track lo, track hi, DMA lo, DMA hi
		mov	hdsk_status, #1				' non-zero status while expecting parameters?
		mov	io_dir, #dir_out
		jmp	#command_0

'**********************************************************************************************
'
' Params expects one more out, the drive
' number, to be written to the port
'
output_hdsk_drive
		mov	hdsk_cmd, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #1			' expect drive number
		mov	io_dir, #dir_out
		jmp	#command_0

'**********************************************************************************************
'
' Params returns 19 bytes on the following
' reads from the port
output_hdsk_params_drive
		cmp	data, #8	WZ, WC			' we support drives 0 to 7
	if_ae	jmp	#:error
		mov	hdsk_params_ptr, hdsk_dpb
		mov	hdsk_params_cnt, #19			' 19 bytes to read
		mov	hdsk_status, #0
		mov	io_dir, #dir_in
		jmp	#command_0
:error
		mov	hdsk_status, #$ff
		jmp	#command_0

'**********************************************************************************************
'
' Wait until the XMM driver clears the
' long[pdo_command]
'
pdo_wait_busy
		wrlong	t1, pdo_command				' send the command
:wait		rdlong	t1, pdo_command
		shl	t1, #1		WZ, WC, NR		' look for negative values or zero
  if_nz_and_nc	jmp	#:wait
	if_nz	mov	hdsk_status, t1
	if_z	mov	hdsk_status, #0
pdo_wait_busy_ret
		ret

'**********************************************************************************************
'
' Setup the DMA address, iobuff pointer and transfer length
'
hdsk_setup_dma
		rdword	t1, hdsk_dma				' get DMA address
#ifdef BANKED_MEM
		cmp	t1, banklimit		WC
	if_c	add	t1, bank				' add the current bank base
#endif
		wrlong	t1, pdo_blocknr				' set DMA address of 128 byte sector
		mov	t1, iobuff_ptr				' 512 byte buffer
		add	t1, sector_offs				' + 128 byte sector offset
		or	t1, h128				' 128 bytes
		wrlong	t1, pdo_address				' as XMM transfer length + hub RAM address
hdsk_setup_dma_ret
		ret

'**********************************************************************************************
'
' Execute the hdsk read
'
hdsk_read_dma
		call	#spi_read_block
		call	#hdsk_setup_dma
		mov	t1, #"W"				' XMM write command
		call	#pdo_wait_busy
		jmp	#command_0

'**********************************************************************************************
'
' Execute the hdsk write
'
hdsk_write_dma
		call	#spi_read_block				' make sure we have the sector in our buffer
		call	#hdsk_setup_dma
		mov	t1, #"R"				' XMM read command
		call	#pdo_wait_busy
#ifdef TIMED_FLUSH
		rdlong	time_dirty, #0				' get clockfrq
		add	time_dirty, CNT
#else
		call	#spi_write_block
#endif !TIMED_FLUSH
		jmp	#command_0

'**********************************************************************************************
'
' Read a logical block address from SD
'
spi_read_block
		rdbyte	lba, hdsk_drive				' get drive number
		and	lba, #7					' we support 8 drives, no more
		shl	lba, #2					' * 4
		add	lba, hdsk_lba_drives			' offset of the per drive LBA0
		rdlong	lba0, lba
		rdword	lba, hdsk_track				' lba = track number
		shl	lba, #5					' * 32 sectors per track
		rdbyte	t1, hdsk_sector				' t1 = sector
		add	lba, t1					' + sector number
		mov	sector_offs, lba
		and	sector_offs, #3				' mask sector offset 0 to 3
		shl	sector_offs, #7				' * 128
		shr	lba, #2					' sector number / 4 (read 512 byte sectors)
		add	lba, lba0				' + first sector of this drive
		and	lba, low_24bits				' make sure bits 31..24 are zero
		cmp	lba, iobuff_lba		WZ		' == buffered lba?
	if_z	jmp	#spi_read_block_ret			' yes, no need to read again
#ifdef TIMED_FLUSH
		cmp	time_dirty, #0		WZ		' flush a pending write
	if_nz	call	#spi_write_block
#endif TIMED_FLUSH
		mov	iobuff_lba, lba
		wrlong	iobuff_lba, pdo_blocknr
		wrlong	iobuff_ptr, pdo_address
		mov	t1, #"r"				' SD/SPI read command
		call	#pdo_wait_busy
spi_read_block_ret
		ret

'**********************************************************************************************
'
' Write a logical block address to SD
'
spi_write_block
		wrlong	iobuff_lba, pdo_blocknr
		wrlong	iobuff_ptr, pdo_address
		mov	t1, #"w"				' SD/SPI write command
		call	#pdo_wait_busy
#ifdef TIMED_FLUSH
		mov	time_dirty, #0				' buffer is clean
#endif
spi_write_block_ret
		ret


'**********************************************************************************************
'
' Read i/o port to read from
'
input
		andn	iocmd, #$ff				' clear the iocmd data bits 7..0
		mov	ptr, input_port_ptr
:loop		rdword	t1, ptr
		add	ptr, #2
		cmp	port, t1			WZ
	if_z	rdword	t1, ptr
	if_z	jmp	t1
		add	ptr, #2
		tjnz	t1, #:loop

		jmp	#command_0

'**********************************************************************************************
'
' Read the console status
'
input_constat
		mov	lmm_pc, #lmm_st0
		call	#lmm_func
	if_z	jmp	#input_keystat
input_lcdstat
input_constat_1
	if_z	mov	data, #%00000010	' no data available
	if_nz	mov	data, #%00000011	' data is available
input_data
		or	iocmd, data
		jmp	#command_0
input_keystat
		mov	lmm_pc, #lmm_keystat
		call	#lmm_func
		jmp	#input_constat_1

input_sio1stat
input_sio2stat
input_sio3stat
input_sio1data
input_sio2data
input_sio3data
input_lcddata
		mov	data, #0
		jmp	#input_data

'**********************************************************************************************
'
' Read the console data
'
' returns 0 if no data is available
'
input_condata
		mov	lmm_pc, #lmm_rx0
		call	#lmm_func
	if_nz	jmp	#input_data

input_keydata
		mov	lmm_pc, #lmm_keydata
		call	#lmm_func
		jmp	#input_data

'**********************************************************************************************
'
' Read the paper punch status
'
input_punstat
		mov	lmm_pc, #lmm_st1	' punch status
		call	#lmm_func
		jmp	#input_constat_1

'**********************************************************************************************
'
' Read the paper punch data
'
' returns 0 if no data is available
'
input_pundata
		mov	lmm_pc, #lmm_rx1
		call	#lmm_func
		jmp	#input_data

'**********************************************************************************************
'
' Output a character on the console (terminal and VT100 emulation, if enabled)
'
' returns 0 if no data is available
'
conout_char
		mov	lmm_pc, #lmm_tx0
		call	#lmm_func
#ifdef HAVE_VGA
		mov	lmm_pc, #lmm_vt100_out
		call	#lmm_func
#endif
#ifdef HAVE_TV
		mov	lmm_pc, #lmm_vt100_out
		call	#lmm_func
#endif
conout_char_ret
		ret

'**********************************************************************************************
'
' Read from the simh port
'
input_simh
		cmp	io_dir, #dir_in		WZ
	if_z	tjnz	simh_params_ptr, #input_simh_params
		or	iocmd, simh_status
		jmp	#command_0

'**********************************************************************************************
'
' Read port value from RAM at simh_params_ptr
'
input_simh_params
		rdbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		sub	simh_params_cnt, #1	WZ
	if_z	mov	simh_params_ptr, #0		' stop replying parameters
		jmp	#input_data

'**********************************************************************************************
'
' Read from the hdsk port
'
input_hdsk
		cmp	io_dir, #dir_in		WZ
	if_z	tjnz	hdsk_params_ptr, #input_hdsk_params
		or	iocmd, hdsk_status
		jmp	#command_0

'**********************************************************************************************
'
' Read port value from RAM at hdsk_params_ptr
'
input_hdsk_params
		rdbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		sub	hdsk_params_cnt, #1	WZ
	if_z	mov	hdsk_params_ptr, #0		' stop replying parameters
		jmp	#input_data

'**********************************************************************************************
'
' Timer interrupt executed at a rate of timer_delta milliseconds
'
timer_int
		add	timer_tick, timer_delta			' schedule next interrupt call
		tjz	irq_ptr, #timer_int_ret			' no IRQ vector set?
		wrlong	timer_addr, irq_ptr
timer_int_ret
		ret

'**********************************************************************************************
'
' Send a text character to the LCD
'
lcd_send_text
		mov	lcd_or, #1
		jmp	#lcd_send

'**********************************************************************************************
'
' Send an instruction to the LCD
'
lcd_send_inst
		mov	lcd_or, #0
lcd_send
		mov	t1, data
		shr	t1, #2
		and	t1, #%00111100
		or	t1, lcd_or
		call	#lcd_pulse_out
		mov	t1, data
		shl	t1, #2
		and	t1, #%00111100
		or	t1, lcd_or
		call	#lcd_pulse_out
lcd_send_ret
lcd_send_inst_ret
lcd_send_text_ret
		ret

'**********************************************************************************************
'
' Clock a 4 bit word in t1 to the LCD
'
lcd_pulse_out
		or	t1, #%0000010		' enable pin high
		call	#lcd_output
		mov	t2, lcd_delay
		add	t2, CNT
		waitcnt	t2, lcd_delay
		andn	t1, #%0000010		' enable pin low
		call	#lcd_output
		waitcnt	t2, lcd_delay
lcd_pulse_out_ret
		ret

'**********************************************************************************************
'
' Output a value to the LCD
'
lcd_output
		mov	dira, latch_dir		' set latch direction
		and	t1, #$ff
		or	t1, latch_lcd		' latch LCD with gate on
		mov	outa, t1
		or	outa, #%1_00000000	' toggle gate on
		mov	dira, #0		' tristate the outputs
lcd_output_ret
		ret

'**********************************************************************************************
'
' LMM function call. lmm_pc contains the function number to execute
'
lmm_func	shl	lmm_pc, #1				' function number * 2
		add	lmm_pc, lmm_func_ptr			' + function table pointer
		rdword	lmm_pc, lmm_pc				' get the new LMM program counter
lmm_next	rdlong	:op1, lmm_pc
		add	lmm_pc, #4
:op1		nop
		rdlong	:op2, lmm_pc
		add	lmm_pc, #4
:op2		nop
		rdlong	:op3, lmm_pc
		add	lmm_pc, #4
:op3		nop
		rdlong	:op4, lmm_pc
		add	lmm_pc, #4
:op4		nop
		jmp	#lmm_next
lmm_func_ret
		ret

		long	0[$1f0-$]

		fit	$1f0

#ifdef HAVE_TV
		long	0[SCREEN_W * SCREEN_H / 4 - $1f0]	' 80x25 = 2000 = $1f4 longs
#endif

#ifdef HAVE_VGA
		long	0[SCREEN_W * SCREEN_H / 4 - $1f0]	' 80x40 = 3200 = $320 longs
vga_vsync	long
vga_cursor	byte	0[8]
vga_colors	word	0[64]					' max 64 rows
#endif HAVE_VGA

output_port_table
		word	condata_port,		output_conout
		word	pundata_port,		output_pundata
		word	lcdstat_port,		output_lcdctrl
		word	lcddata_port,		output_lcddata
		word	hdsk_port,		output_hdsk
		word	simh_port,		output_simh
		word	0,			command_0

input_port_table
		word	simh_port,		input_simh
		word	hdsk_port,		input_hdsk
		word	constat_port,		input_constat
		word	condata_port,		input_condata
		word	punstat_port,		input_punstat
		word	pundata_port,		input_pundata
		word	sio1stat_port,		input_sio1stat
		word	sio1data_port,		input_sio1data
		word	sio2stat_port,		input_sio2stat
		word	sio2data_port,		input_sio2data
		word	sio3stat_port,		input_sio3stat
		word	sio3data_port,		input_sio3data
		word	lcdstat_port,		input_lcdstat
		word	lcddata_port,		input_lcddata
		word	keystat_port,		input_keystat
		word	keydata_port,		input_keydata
		word	0,			command_0

simh_cmd_table
		word	simh_set_bank_select,	output_simh_set_bank_select
		word	simh_get_common,	output_simh_get_common
		word	simh_reset,		output_simh_reset
		word	simh_has_banked_memory,	output_simh_has_banked_memory
		word	simh_start_timer,	output_simh_start_timer
		word	simh_stop_timer,	output_simh_stop_timer
		word	simh_timer_delta,	output_simh_timer_delta
		word	simh_timer_addr,	output_simh_timer_addr
		word	0,			command_0

lmm_func_table
		word	@@@mul32				' multiply x by y
		word	@@@div32				' divide x by y
		word	@@@break				' break from Z80
		word	@@@keystat				' read keyboard status
		word	@@@keydata				' read keyboard data
		word	@@@st0					' status of SIO port #0 input
		word	@@@tx0					' send a character on SIO port #0
		word	@@@rx0					' receive a character on SIO port #0
		word	@@@st1					' status of SIO port #1 input
		word	@@@tx1					' send a character on SIO port #1
		word	@@@rx1					' receive a character on SIO port #1
#ifdef HAVE_TV
		word	@@@vt100_out				' 2 - send a character to the VT100 emulation
#elseifdef HAVE_VGA
		word	@@@vt100_out				' 2 - send a character to the VT100 emulation
#else
		word	command					' 2 - unused
#endif HAVE_VGA
#ifdef HAVE_LCD
		word	@@@lcd_init				' initialize LCD
#else
		word	command					' 2 - unused
#endif

'**********************************************************************************************
'
'	LMM CODE FRAGMENTS
'
'**********************************************************************************************


'**********************************************************************************************
'
' Multiply x by y 32 bit, result low in x, result high in t1
'
mul32
		mov	t1, #0
		mov	t2, #32
		shr     x, #1		WC
:loop	if_c	add	t1, y		WC
		rcr	t1, #1		WC
		rcr	x, #1		WC
		sub	t2, #1		WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :loop)
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Divide x by y 32 bit, result low in x, remainder in y
'
div32
		mov	t1, #0
		mov	t2, #32
:loop1		shr	y, #1		WZ, WC
		rcr	t1, #1
	if_nz	sub	t2, #1		WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :loop1)
:loop2		cmpsub	x, t1		WC
		rcl	y, #1
		shr	t1, #1
		sub	t2, #1		WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :loop2)
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
		mov	t1, iocmd			' read the address of the register file
		shr	t1, #8				' bits 23..8 to 15..0
		and	t1, low_16bits			' mask only bits 15..0
		mov	ptr, msg_break_ptr	

		mov	t2, #8
:loop_1		call	#conout_string
		rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
		call	#conout_hex_w
		add	t1, #2
		sub	t2, #1			WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :loop_1)

		call	#conout_string			' new line and print "R:"
		rdbyte	hexval, t1			' R
		add	t1, #1
		rdbyte	t2, t1				' R2
		and	hexval, #%01111111		' R register's bit #7
		or	hexval, t2			' does never change from the last written value
		call	#conout_hex_b
		add	t1, #1

		mov	t2, #3
		call	#conout_string
		rdbyte	hexval, t1			' IFF1+2/IM
		mov	t2, t1				' save value
		and	hexval, #1			' IFF1 (interrupt flip-flop 1)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #1			' shift bit #1 to #0
		and	hexval, #1			' IFF2 (interrupt flip-flop 2)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #2			' shift bits #3..2 to #1..0
		and	hexval, #3			' IM (interrupt mode)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' I (interrupt vector register)
		call	#conout_hex_b
		add	t1, #1

		mov	t2, #4
:loop_2
		call	#conout_string
		rdword	hexval, t1			' IX, IY, SP, PC
		mov	t3, t1
		call	#conout_hex_w
		add	t1, #2
		sub	t2, #1			WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :loop_2)

		call	#conout_string

		mov	t2, #4				' 4 bytes
		shl	t2, #16
		or	t2, t3				' length in upper 16 bits, address in lower 16 bits
		wrlong	t2, pdo_address			' as XMM ram address
		mov	t2, #"R"			' XMM read command
		wrlong	t2, pdo_command
:wait		rdlong	t2, pdo_command		WZ, WC
  if_nc_and_nz	sub	lmm_pc, #4*($ + 1 - :wait)

		rdbyte	hexval, t1
		call	#conout_hex_b
		call	#conout_crlf
		jmp	#lmm_func_ret

#ifdef HAVE_KBD
'**********************************************************************************************
'
' Read the keyboard status
'
keystat
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
		cmp	t1, t2			WZ
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Read the keyboard data
'
keydata
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
		mov	t3, t2
		cmp	t1, t2		WZ
		mov	data, #0
	if_z	jmp	#lmm_func_ret

		shl	t2, #1
		add	t2, kbd_keys
		rdword	data, t2		' read just the keyboard word
		add	t3, #1
		and	t3, #15
		wrlong	t3, kbd_tail
	
		' process key codes
		test	data, kbd_ctrl		WZ	' is it a control code?
	if_nz	and	data, #$1f			' make it $00..$1f
	if_nz	jmp	#lmm_func_ret
		cmp	data, #200		WZ	' backspace?
	if_z	mov	data, #8			' backspace character
	if_z	jmp	#lmm_func_ret
		cmp	data, #201		WZ	' is it delete?
	if_z	mov	data, #127			' Delete character
	if_z	jmp	#lmm_func_ret
		cmp	data, #203		WZ	' upper code for Esc?
	if_z	mov	data, #27			' Escape character
	if_z	jmp	#lmm_func_ret
		cmp	data, #128		WZ, WC	' don't return other codes
	if_ae	mov	data, #0
#else
keystat
keydata
		mov	data, #0		WZ
		jmp	#lmm_func_ret
#endif HAVE_KBD

'**********************************************************************************************
'
' Read status for the FullDuplexSerial rx_buffer0
'
st0
		mov	data, #0				' no data available
		rdlong	t1, rx_head0
		rdlong	t2, rx_tail0
		cmp	t1, t2		WZ			' set zero flag if no data available
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Write to the FullDuplexSerial tx_buffer0
'
tx0
:wait
		rdlong	sio_head, tx_head0
		mov	sio_temp, sio_head
		add	sio_head, #1
		rdlong	sio_tail, tx_tail0
		and	sio_head, #15
		cmp	sio_head, sio_tail	WZ
	if_z	sub	lmm_pc, #4*($ + 1 - :wait)		' wait until there's room in the buffer
		mov	addr, tx_buffer0
		add	addr, sio_temp
		wrbyte	data, addr
		wrlong	sio_head, tx_head0			' increment the tx_head0
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Read from the FullDuplexSerial rx_buffer0
'
rx0
		mov	data, #0				' no data available
		rdlong	t1, rx_head0
		rdlong	t2, rx_tail0
		cmp	t1, t2		WZ
	if_z	jmp	#lmm_func_ret
		mov	addr, rx_buffer0
		add	addr, t2
		rdbyte	data, addr	WZ
		add	t2, #1
		and	t2, #63
		wrlong	t2, rx_tail0
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Read status for the FullDuplexSerial rx_buffer1
'
st1
		mov	data, #0				' no data available
		rdlong	t1, rx_head1
		rdlong	t2, rx_tail1
		cmp	t1, t2		WZ			' set zero flag if no data available
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Write to the FullDuplexSerial tx_buffer1
'
tx1
:wait
		rdlong	sio_head, tx_head1
		mov	sio_temp, sio_head
		add	sio_head, #1
		rdlong	sio_tail, tx_tail1
		and	sio_head, #15
		cmp	sio_head, sio_tail	WZ
	if_z	sub	lmm_pc,#4*($ + 1 - :wait)		' wait until there's room in the buffer
		mov	addr, tx_buffer1
		add	addr, sio_temp
		wrbyte	data, addr
		wrlong	sio_head, tx_head1			' increment the tx_head1
		jmp	#lmm_func_ret

'**********************************************************************************************
'
' Read from the FullDuplexSerial rx_buffer1
'
rx1
		mov	data, #0			' no data available
		rdlong	t1, rx_head1
		rdlong	t2, rx_tail1
		cmp	t1, t2			WZ
	if_z	jmp	#lmm_func_ret
		mov	addr, rx_buffer1
		add	addr, t2
		rdbyte	data, addr		WZ
		add	t2, #1
		and	t2, #63
		wrlong	t2, rx_tail1
		jmp	#lmm_func_ret

#ifdef HAVE_VGA
'**********************************************************************************************
'
' Write to the VGA rendezvous long
'
vt100_out
:wait		rdlong	sio_temp, vt_cmdptr	WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :wait)
		mov	sio_temp, #$100
		or	sio_temp, data
		wrlong	sio_temp, vt_cmdptr
		jmp	#lmm_func_ret
#endif HAVE_VGA

#ifdef HAVE_TV
'**********************************************************************************************
'
' Write to the VT100 on TV rendezvous long
'
vt100_out
:wait		rdlong	sio_temp, vt_cmdptr	WZ
	if_nz	sub	lmm_pc, #4*($ + 1 - :wait)
		mov	sio_temp, #$100
		or	sio_temp, data
		wrlong	sio_temp, vt_cmdptr
		jmp	#lmm_func_ret
#endif HAVE_TV

#ifdef HAVE_LCD
'**********************************************************************************************
'
' Initialize the LC display
'
lcd_init
		mov	t1, #%00001100			' send out 0011 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	t1, #%00001100			' send out 0011 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	t1, #%00001100			' send out 0011 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	t1, #%00001000			' send out 0010 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	t1, #%00001000			' send out 0010 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	t1, #%00100000			' send out 1000 already shifted to xxnnnnxx
		call	#lcd_pulse_out
		mov	data, #14			' cursor on
		call	#lcd_send_inst
		mov	data, #1			' clear screen
		call	#lcd_send_inst
		mov	data, #13			' flashing cursor
		call	#lcd_send_inst
		mov	data, #$80			' to position 1
		call	#lcd_send_inst
lcd_init_ret
		jmp	#lmm_func_ret
#endif HAVE_LCD

drive_name	long	@@@drive_name_a, @@@drive_name_b, @@@drive_name_c, @@@drive_name_d
		long	@@@drive_name_e, @@@drive_name_f, @@@drive_name_g, @@@drive_name_h

simh_params_	byte	0,0,0,0,0,0,0,0		' parameters of command

lba_drive_a	long	0			' first sector of drive A:
lba_drive_b	long	0			' first sector of drive B:
lba_drive_c	long	0			' first sector of drive C:
lba_drive_d	long	0			' first sector of drive D:
lba_drive_e	long	0			' first sector of drive E:
lba_drive_f	long	0			' first sector of drive F:
lba_drive_g	long	0			' first sector of drive G:
lba_drive_h	long	0			' first sector of drive H:

hdsk_params_	byte
hdsk_drive_	byte	0
hdsk_sector_	byte	0
hdsk_track_	byte	0, 0
hdsk_dma_	byte	0, 0
hdsk_dummy_	word	0

'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb_	byte			' 8MByte Default SIMH Altair HDSK params
hdsk_spt_lo	byte	$20		' sectors per track (low byte)
hdsk_spt_hi	byte	$00		' sectors per track (high byte)
hdsk_bsh	byte	$05		' data allocation Block SHift factor
hdsk_blm	byte	$1F		' data allocation block mask
hdsk_exm	byte	$01		' EXtent Mask
hdsk_dsm_lo	byte	$F9		' maximum data block number (low_byte)
hdsk_dsm_hi	byte	$07		' maximum data block number (high_byte)
hdsk_drm_lo	byte	$FF		' total number of directory entries (low byte)
hdsk_drm_hi	byte	$03		' total number of directory entries (high byte)
hdsk_al0	byte	$FF		' determine reserved directory blocks
hdsk_al1	byte	$00		' determine reserved directory blocks
hdsk_cks_lo	byte	$00		' size of directory ChecK vector (low byte)
hdsk_cks_hi	byte	$00		' size of directory ChecK vector (high byte)
hdsk_off_lo	byte	$06		' number of reserved tracks (offset) (low byte)
hdsk_off_hi	byte	$00		' number of reserved tracks (offset) (high byte)
hdsk_psh	byte	$00		' Physical record SHift factor, CP/M 3
hdsk_phm	byte	$00		' Physical record Mask, CP/M 3
hdsk_ss_lo	byte	$80		' Sector Size (low byte)
hdsk_ss_hi	byte	$00		' Sector Size (high byte)
hdsk_filler	byte	$00		' align to word boundary

#ifdef FAT12_SUPPORT
iobuff		long	0[2*512/4]	' I/O buffer (need 2 because of FAT12)
#else
iobuff		long	0[512/4]	' I/O buffer for one 512 byte sector
#endif

boot_name	byte	"BOOT.DSK",0
drive_name_a	byte	"A.DSK",0
drive_name_b	byte	"B.DSK",0
drive_name_c	byte	"C.DSK",0
drive_name_d	byte	"D.DSK",0
drive_name_e	byte	"E.DSK",0
drive_name_f	byte	"F.DSK",0
drive_name_g	byte	"G.DSK",0
drive_name_h	byte	"H.DSK",0

msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" BC'",0
		byte	" DE'",0
		byte	" HL'",0
		byte	" AF'",0
		byte	13,10, "R:",0
		byte	" IFF1:",0
		byte	" IFF2:",0
		byte	" IM:",0
		byte	" I:",0
		byte	" IX:",0
		byte	" IY:",0
		byte	" SP:",0
		byte	" PC:",0
		byte	" = ",0
